from mahjong_utils.lib import libmahjongutils


def get_parent_point_by_han_hu(han: int, hu: int):
    """
    获取亲家X番Y符的点数

    :param han: 番
    :param hu: 符
    :return: (荣和点数, 自摸各家点数)
    """
    result = libmahjongutils.call("getParentPointByHanHu", {
        "han": han,
        "hu": hu
    })

    return result["ron"], result["tsumo"]


def get_child_point_by_han_hu(han: int, hu: int):
    """
    获取子家X番Y符的点数

    :param han: 番
    :param hu: 符
    :return: (荣和点数, 自摸庄家点数, 自摸闲家点数)
    """
    result = libmahjongutils.call("getChildPointByHanHu", {
        "han": han,
        "hu": hu
    })

    return result["ron"], result["tsumoParent"], result["tsumoChild"]


__all__ = ("get_parent_point_by_han_hu", "get_child_point_by_han_hu",)



from typing import Optional, Set, List
from typing import Tuple

from pydantic import BaseModel
from stringcase import snakecase, pascalcase

from mahjong_utils.lib import libmahjongutils
from mahjong_utils.models.furo import Furo
from mahjong_utils.models.hora_hand_pattern import HoraHandPattern
from mahjong_utils.models.tile import Tile
from mahjong_utils.models.wind import Wind
from mahjong_utils.point_by_han_hu import get_parent_point_by_han_hu, get_child_point_by_han_hu
from mahjong_utils.shanten import ShantenResult
from mahjong_utils.yaku import Yaku, get_yaku


class Hora(BaseModel):
    pattern: HoraHandPattern
    han: int
    dora: int
    yaku: Set[Yaku]
    extra_yaku: Set[Yaku]
    has_yakuman: bool

    @classmethod
    def decode(cls, data: dict) -> "Hora":
        return Hora(
            pattern=HoraHandPattern.__decode__(data["pattern"]),
            han=data["han"],
            dora=data["dora"],
            yaku=set(get_yaku(snakecase(yk)) for yk in data["yaku"]),
            extra_yaku=set(get_yaku(snakecase(yk)) for yk in data["extraYaku"]),
            has_yakuman=data["hasYakuman"]
        )

    @property
    def hu(self) -> int:
        return self.pattern.hu

    @property
    def tsumo(self) -> bool:
        return self.pattern.tsumo

    @property
    def self_wind(self) -> Optional[Wind]:
        return self.pattern.self_wind

    @property
    def round_wind(self) -> Optional[Wind]:
        return self.pattern.round_wind

    @property
    def agari(self) -> Tile:
        return self.pattern.agari

    @property
    def parent_point(self) -> Tuple[int, int]:
        """
        亲家和牌点数

        :return: (荣和点数, 自摸各家点数)
        """
        if len(self.yaku) == 0:
            return 0, 0
        elif self.has_yakuman:
            times = 0
            for yaku in self.yaku:
                times += yaku.han // 13

            ans = get_parent_point_by_han_hu(13, 20)
            return ans[0] * times, ans[1] * times
        else:
            return get_parent_point_by_han_hu(self.han, self.hu)

    @property
    def child_point(self) -> Tuple[int, int, int]:
        """
        子家X番Y符的点数

        :return: (荣和点数, 自摸庄家点数, 自摸闲家点数)
        """
        if len(self.yaku) == 0:
            return 0, 0, 0
        elif self.has_yakuman:
            times = 0
            for yaku in self.yaku:
                times += yaku.han // 13

            ans = get_child_point_by_han_hu(13, 20)
            return ans[0] * times, ans[1] * times, ans[2] * times
        else:
            return get_child_point_by_han_hu(self.han, self.hu)


def build_hora(
        tiles: List[Tile], furo: Optional[List[Furo]], agari: Tile,
        tsumo: bool,
        *, dora: int = 0,
        self_wind: Optional[Wind] = None, round_wind: Optional[Wind] = None,
        extra_yaku: Optional[Set[Yaku]] = None
) -> Hora:
    """
    根据手牌构造Hora

    :param tiles: 手牌
    :param furo: 副露
    :param agari: 和牌
    :param tsumo: 是否自摸
    :param dora: 宝牌数
    :param self_wind: 自风
    :param round_wind: 场风
    :param extra_yaku: 额外役
    :return: Hora
    """
    result = libmahjongutils.call("hora", {
        "tiles": [str(t) for t in tiles],
        "furo": [fr.__encode__() for fr in furo] if furo is not None else [],
        "agari": str(agari),
        "tsumo": tsumo,
        "dora": dora,
        "selfWind": pascalcase(self_wind.name) if self_wind is not None else None,
        "roundWind": pascalcase(round_wind.name) if round_wind is not None else None,
        "extraYaku": [pascalcase(yk.name) for yk in extra_yaku] if extra_yaku is not None else []
    })

    return Hora.decode(result)


def build_hora_from_shanten_result(
        shanten_result: ShantenResult,
        agari: Tile,
        tsumo: bool,
        *, dora: int = 0,
        self_wind: Optional[Wind] = None, round_wind: Optional[Wind] = None,
        extra_yaku: Optional[Set[Yaku]] = None
) -> Hora:
    """
    根据向听分析结果构造Hora

    :param shanten_result: 向听分析结果
    :param agari: 和牌
    :param tsumo: 是否自摸
    :param dora: 宝牌数
    :param self_wind: 自风
    :param round_wind: 场风
    :param extra_yaku: 额外役
    :return: Hora
    """
    result = libmahjongutils.call("hora", {
        "shantenResult": shanten_result.__encode__(),
        "agari": str(agari),
        "tsumo": tsumo,
        "dora": dora,
        "selfWind": pascalcase(self_wind.name) if self_wind is not None else None,
        "roundWind": pascalcase(round_wind.name) if round_wind is not None else None,
        "extraYaku": [pascalcase(yk.name) for yk in extra_yaku] if extra_yaku is not None else []
    })

    return Hora.decode(result)


__all__ = ("Hora", "build_hora", "build_hora_from_shanten_result")



from .common import all_common_yaku
from .extra import all_extra_yaku
from .yaku import Yaku
from .yakuman import all_yakuman

all_yaku = all_common_yaku | all_extra_yaku | all_yakuman

_all_yaku_mapping = dict(map(lambda x: (x.name, x), all_yaku))


def get_yaku(name: str) -> Yaku:
    yaku = _all_yaku_mapping.get(name, None)
    if yaku is None:
        raise ValueError(f"{name} is not a yaku")
    return yaku


__all__ = ("Yaku", "all_yaku", "get_yaku")



import json
import sys
import threading
from importlib import resources
from typing import Optional, Mapping, Any

import cffi


class LibMahjongUtils:
    def __init__(self) -> None:
        self.ffi = cffi.FFI()

        with resources.open_text(__name__, "libmahjongutils_api.i") as f:
            h = f.read()
            self.ffi.cdef(h)

        if sys.platform == 'win32':
            libname = "libmahjongutils.dll"  # windows
        elif sys.platform == 'darwin':
            libname = "libmahjongutils.dylib"  # macOS
        else:
            libname = "libmahjongutils.so"  # unix/linux

        with resources.path(__name__, libname) as libpath:
            self.lib = self.ffi.dlopen(str(libpath))

        self._lib_sy = threading.local()
        self._lib_sy.x = self.lib.libmahjongutils_symbols()

    @property
    def lib_sy(self):
        return self._lib_sy.x

    def call(self, name: str, params: dict,
             params_dumps_kwargs: Optional[Mapping[str, Any]] = None,
             result_loads_kwargs: Optional[Mapping[str, Any]] = None) -> dict:
        if params_dumps_kwargs is None:
            params_dumps_kwargs = {}
        if result_loads_kwargs is None:
            result_loads_kwargs = {}

        params = json.dumps(params, **params_dumps_kwargs)

        entry = self.lib_sy.kotlin.root.mahjongutils.get_ENTRY()
        result = self.lib_sy.kotlin.root.mahjongutils.Entry.call(
            entry,
            self.ffi.new("char[]", name.encode()),
            self.ffi.new("char[]", params.encode()))
        result = self.ffi.string(result)

        result = json.loads(result, **result_loads_kwargs)

        if result['code'] == 200:
            return result['data']
        elif result['code'] == 404:
            raise ValueError(result['msg'])
        elif result['code'] == 400:
            raise ValueError(result['msg'])
        else:
            raise RuntimeError(result['msg'])

    def close(self):
        self.ffi.dlclose(self.lib)


libmahjongutils = LibMahjongUtils()

__all__ = ("LibMahjongUtils", "libmahjongutils")



from abc import ABC, abstractmethod
from typing import Optional

from stringcase import pascalcase, snakecase

from mahjong_utils.models.hand_pattern import RegularHandPattern, HandPattern, ChitoiHandPattern, KokushiHandPattern
from mahjong_utils.models.tatsu import Tatsu
from mahjong_utils.models.tile import Tile
from mahjong_utils.models.wind import Wind


class HoraHandPattern(HandPattern, ABC):
    agari: Tile
    tsumo: bool
    hu: int
    self_wind: Optional[Wind]
    round_wind: Optional[Wind]

    @abstractmethod
    def __encode__(self) -> dict:
        raise NotImplementedError()

    @classmethod
    def __decode__(cls, data: dict) -> "HoraHandPattern":
        if data['type'] == 'RegularHoraHandPattern':
            return RegularHoraHandPattern.__decode__(data)
        elif data['type'] == 'ChitoiHoraHandPattern':
            return ChitoiHoraHandPattern.__decode__(data)
        elif data['type'] == 'KokushiHoraHandPattern':
            return KokushiHoraHandPattern.__decode__(data)
        else:
            raise ValueError("invalid type: " + data['type'])


class RegularHoraHandPattern(HoraHandPattern, RegularHandPattern):
    agari_tatsu: Optional[Tatsu]
    jyantou: Tile

    def __encode__(self) -> dict:
        return dict(
            type="RegularHoraHandPattern",
            agari=self.agari.__encode__(),
            tsumo=self.tsumo,
            self_wind=pascalcase(self.self_wind.name) if self.self_wind is not None else None,
            round_wind=pascalcase(self.round_wind.name) if self.round_wind is not None else None,
            agari_tatsu=self.agari_tatsu.__encode__() if self.agari_tatsu is not None else None,
            pattern=RegularHandPattern.__encode__(self)
        )

    @classmethod
    def __decode__(cls, data: dict) -> "RegularHoraHandPattern":
        return RegularHoraHandPattern(
            agari=Tile.__decode__(data["agari"]),
            tsumo=data["tsumo"],
            self_wind=Wind[snakecase(data["selfWind"])] if data["selfWind"] is not None else None,
            round_wind=Wind[snakecase(data["roundWind"])] if data["roundWind"] is not None else None,
            agari_tatsu=Tatsu.__decode__(data["agariTatsu"]) if data["agariTatsu"] is not None else None,
            hu=data["hu"],
            **RegularHandPattern.__decode__(data["pattern"]).dict()
        )


class ChitoiHoraHandPattern(HoraHandPattern, ChitoiHandPattern):
    hu: int = 25

    def __encode__(self) -> dict:
        return dict(
            type="ChitoiHoraHandPattern",
            agari=self.agari.__encode__(),
            tsumo=self.tsumo,
            self_wind=pascalcase(self.self_wind.name) if self.self_wind is not None else None,
            round_wind=pascalcase(self.round_wind.name) if self.round_wind is not None else None,
            pattern=ChitoiHandPattern.__encode__(self)
        )

    @classmethod
    def __decode__(cls, data: dict) -> "ChitoiHoraHandPattern":
        return ChitoiHoraHandPattern(
            agari=Tile.__decode__(data["agari"]),
            tsumo=data["tsumo"],
            self_wind=Wind[snakecase(data["selfWind"])] if data["selfWind"] is not None else None,
            round_wind=Wind[snakecase(data["roundWind"])] if data["roundWind"] is not None else None,
            **ChitoiHandPattern.__decode__(data["pattern"]).dict()
        )


class KokushiHoraHandPattern(HoraHandPattern, KokushiHandPattern):
    hu: int = 20
    repeated: Tile

    @property
    def thirteen_waiting(self) -> bool:
        return self.agari == self.repeated

    def __encode__(self) -> dict:
        return dict(
            type="KokushiHoraHandPattern",
            agari=self.agari.__encode__(),
            tsumo=self.tsumo,
            self_wind=pascalcase(self.self_wind.name) if self.self_wind is not None else None,
            round_wind=pascalcase(self.round_wind.name) if self.round_wind is not None else None,
            pattern=KokushiHandPattern.__encode__(self)
        )

    @classmethod
    def __decode__(cls, data: dict) -> "KokushiHoraHandPattern":
        return KokushiHoraHandPattern(
            agari=Tile.__decode__(data["agari"]),
            tsumo=data["tsumo"],
            self_wind=Wind[snakecase(data["selfWind"])] if data["selfWind"] is not None else None,
            round_wind=Wind[snakecase(data["roundWind"])] if data["roundWind"] is not None else None,
            **KokushiHandPattern.__decode__(data["pattern"]).dict()
        )



from abc import ABC, abstractmethod
from typing import List, Sequence, Union, Iterable, TYPE_CHECKING

from pydantic.dataclasses import dataclass

from .tile import Tile, parse_tiles
from .tile_type import TileType

if TYPE_CHECKING:
    from .tatsu import Penchan, Kanchan, Ryanmen, Tatsu, Toitsu


@dataclass(frozen=True)
class Mentsu(ABC):
    """
    面子
    """
    tile: Tile

    @property
    @abstractmethod
    def tiles(self) -> Iterable[Tile]:
        raise NotImplementedError()

    @abstractmethod
    def after_discard(self, discard: Tile) -> "Tatsu":
        raise NotImplementedError()

    def __encode__(self) -> dict:
        return dict(type=type(self).__name__, tile=self.tile.__encode__())

    @classmethod
    def __decode__(cls, data: dict) -> "Mentsu":
        t = Tile.__decode__(data["tile"])
        if data['type'] == 'Kotsu':
            return Kotsu(t)
        elif data['type'] == 'Shuntsu':
            return Shuntsu(t)
        else:
            raise ValueError("invalid type: " + data['type'])

    @staticmethod
    def parse(t: Union[Sequence[Tile], str]) -> "Mentsu":
        if isinstance(t, str):
            t = parse_tiles(t)

        if len(t) != 3:
            raise ValueError("_tiles must has length of 3")

        if t[0] == t[1] == t[2]:
            return Kotsu(t[0])
        else:
            if t[0].tile_type == TileType.Z or t[1].tile_type == TileType.Z or t[2].tile_type == TileType.Z:
                raise ValueError(f"invalid _tiles: {t}")
            t = sorted(t)
            if t[1] - t[0] == 1 and t[2] - t[1] == 1:
                return Shuntsu(t[0])
            else:
                raise ValueError(f"invalid _tiles: {t}")


@dataclass(frozen=True)
class Kotsu(Mentsu):
    """
    刻子
    """

    def __repr__(self):
        return f"{self.tile.num}{self.tile.num}{self.tile.num}{self.tile.tile_type}"

    @property
    def tiles(self) -> List[Tile]:
        return [self.tile] * 3

    def after_discard(self, discard: Tile) -> "Toitsu":
        from .tatsu import Toitsu
        if discard == self.tile:
            return Toitsu(discard)
        raise ValueError()


@dataclass(frozen=True)
class Shuntsu(Mentsu):
    """
    顺子
    """

    def __post_init__(self):
        assert self.tile.tile_type != TileType.Z
        assert 1 <= self.tile.num <= 7

    def __repr__(self):
        return f"{self.tile.num}{self.tile.num + 1}{self.tile.num + 2}{self.tile.tile_type}"

    @property
    def tiles(self) -> List[Tile]:
        return [self.tile, self.tile + 1, self.tile + 2]

    def after_discard(self, discard: Tile) -> "Union[Penchan, Ryanmen, Kanchan]":
        from .tatsu import Penchan, Kanchan, Ryanmen
        if discard == self.tile:
            if discard.num == 7:
                return Penchan(discard + 1)
            elif discard.num == 3:
                return Penchan(discard - 2)
            else:
                return Ryanmen(discard + 1)
        elif discard == self.tile + 1:
            return Kanchan(self.tile)
        elif discard == self.tile + 2:
            if self.tile.num == 1:
                return Penchan(self.tile)
            else:
                return Ryanmen(self.tile)
        raise ValueError()


__all__ = ("Mentsu", "Kotsu", "Shuntsu",)



from enum import Enum


class TileType(str, Enum):
    """
    牌的花色
    """

    M = "m"
    P = "p"
    S = "s"
    Z = "z"

    @property
    def num_range(self):
        if self == TileType.Z:
            return range(1, 8)
        else:
            return range(0, 10)

tile_type_index_mapping = {TileType.M: 0, TileType.P: 1, TileType.S: 2, TileType.Z: 3}
tile_type_reversed_index_mapping = [TileType.M, TileType.P, TileType.S, TileType.Z]

__all__ = ("TileType","tile_type_index_mapping", "tile_type_reversed_index_mapping")



from . import furo
from . import hand
from . import hand_pattern
from . import hora_hand_pattern
from . import mentsu
from . import shanten
from . import tatsu
from . import tile
from . import tile_type
from . import wind



from io import StringIO
from typing import Union, List, Iterable, Optional, NamedTuple

from mahjong_utils.models.tile_type import TileType, tile_type_index_mapping, tile_type_reversed_index_mapping


class Tile(NamedTuple):
    """
    牌
    """

    tile_type: TileType
    num: int

    def __repr__(self):
        return f"{self.num}{self.tile_type}"

    @property
    def real_num(self) -> int:
        if self.tile_type != TileType.Z and self.num == 0:
            return 5
        else:
            return self.num

    @property
    def code(self):
        return tile_type_index_mapping[self.tile_type] * 10 + self.num

    def __hash__(self):
        return self.code

    def __cmp__(self, other):
        if not isinstance(other, Tile):
            raise TypeError(other)

        if self.tile_type != other.tile_type:
            return ord(self.tile_type) - ord(other.tile_type)
        elif self.num != 0 and other.num != 0:
            # 两张都不是红dora
            return self.num - other.num
        elif self.num == 0 and other.num == 0:
            # 两张都是红dora
            return 0
        elif self.num == 0:
            # 这张是红dora
            if other.num > 5:
                return -1
            else:
                return 1
        else:
            # 另一张是红dora
            if self.num <= 5:
                return -1
            else:
                return 1

    def __eq__(self, other):
        return other is not None and self.__cmp__(other) == 0

    def __ne__(self, other):
        return self.__cmp__(other) != 0

    def __gt__(self, other):
        return self.__cmp__(other) > 0

    def __lt__(self, other):
        return self.__cmp__(other) < 0

    def __ge__(self, other):
        return self.__cmp__(other) >= 0

    def __le__(self, other):
        return self.__cmp__(other) <= 0

    def __add__(self, other):
        pending_num = self.num + other
        if pending_num in self.tile_type.num_range:
            return self.by_type_and_num(self.tile_type, pending_num)
        else:
            raise ValueError("num out of range")

    def __sub__(self, other):
        if isinstance(other, int):
            return self + (-other)
        elif isinstance(other, Tile):
            if self.tile_type != other.tile_type:
                raise ValueError(f"tile_type of {self} and {other} are not the same")
            return self.num - other.num

    @classmethod
    def by_type_and_num(cls, tile_type: TileType, num: int) -> "Tile":
        tile_type_index = tile_type_index_mapping[tile_type]

        if 0 <= tile_type_index * 10 + num < len(tile_pool):
            t = tile_pool[tile_type_index * 10 + num]
            if t is not None:
                return t

        raise ValueError(f"invalid arguments: tile_type={tile_type}, num={num}")

    @classmethod
    def by_text(cls, text: str) -> "Tile":
        if len(text) != 2:
            raise ValueError(f"invalid arguments: text={text}")

        if text[1].lower() == 'm':
            tile_type = TileType.M
        elif text[1].lower() == 's':
            tile_type = TileType.S
        elif text[1].lower() == 'p':
            tile_type = TileType.P
        elif text[1].lower() == 'z':
            tile_type = TileType.Z
        else:
            raise ValueError(f"invalid arguments: text={text}")

        try:
            num = int(text[0])
        except ValueError as e:
            raise ValueError(f"invalid arguments: text={text}") from e

        return cls.by_type_and_num(tile_type, num)

    @classmethod
    def by_code(cls, code: int) -> "Tile":
        tile_type = tile_type_reversed_index_mapping[code // 10]
        num = code % 10
        return cls.by_type_and_num(tile_type, num)

    def __encode__(self) -> str:
        return str(self)

    @classmethod
    def __decode__(cls, data: str) -> "Tile":
        return cls.by_text(data)


tile_pool: List[Optional[Tile]] = []

for i in range(0, 10):
    tile_pool.append(Tile(TileType.M, i))
for i in range(0, 10):
    tile_pool.append(Tile(TileType.P, i))
for i in range(0, 10):
    tile_pool.append(Tile(TileType.S, i))
tile_pool.append(None)
for i in range(1, 8):
    tile_pool.append(Tile(TileType.Z, i))

Tile._tile_pool = tile_pool


def parse_tiles(text: str) -> List[Tile]:
    ans: List[Tile] = []
    pending: List[int] = []
    for c in text:
        if c.upper() in ("M", "P", "S", "Z"):
            tile_type = TileType[c.upper()]
            for num in pending:
                ans.append(Tile.by_type_and_num(tile_type, num))
            pending.clear()
        elif c.isdigit():
            pending.append(ord(c) - ord('0'))
        else:
            raise ValueError(f"invalid character: {c}")

    if len(pending) > 0:
        raise ValueError("missing tile type at the end of your given text")

    return ans


def tiles_text(tiles: Union[Tile, Iterable[Tile]]) -> str:
    if isinstance(tiles, Tile):
        return str(tiles)

    prev = None
    with StringIO() as sio:
        for t in tiles:
            if prev and prev.tile_type != t.tile_type:
                sio.write(prev.tile_type)
            sio.write(str(t.num))
            prev = t

        if prev:
            sio.write(prev.tile_type)

        return sio.getvalue()


all_yaochu = frozenset(parse_tiles("19m19s19p1234567z"))


def is_m(t: Tile) -> bool:
    """
    判断该牌是否为万子
    """
    return t.tile_type == TileType.M


def is_p(t: Tile) -> bool:
    """
    判断该牌是否为筒子
    """
    return t.tile_type == TileType.P


def is_s(t: Tile) -> bool:
    """
    判断该牌是否为索子
    """
    return t.tile_type == TileType.S


def is_z(t: Tile) -> bool:
    """
    判断该牌是否为字牌
    """
    return t.tile_type == TileType.Z


def is_yaochu(t: Tile) -> bool:
    """
    判断该牌是否为幺九牌
    """
    return t in all_yaochu


def is_wind(t: Tile) -> bool:
    """
    判断该牌是否为风牌
    """
    return t.tile_type == TileType.Z and 1 <= t.num <= 4


def is_sangen(t: Tile) -> bool:
    """
    判断该牌是否为三元牌
    """
    return t.tile_type == TileType.Z and 5 <= t.num <= 7


__all__ = ("Tile", "parse_tiles", "tiles_text", "all_yaochu",
           "is_m", "is_p", "is_s", "is_z", "is_wind", "is_sangen", "is_yaochu")



from abc import ABC
from typing import List, Sequence, Union

from pydantic.dataclasses import dataclass

from .mentsu import Shuntsu, Kotsu, Mentsu
from .tile import Tile, parse_tiles
from .tile_type import TileType


class Furo(Mentsu, ABC):
    """
    副露
    """

    def __encode__(self) -> dict:
        return dict(type=type(self).__name__, tile=self.tile.__encode__())

    @classmethod
    def __decode__(cls, data: dict) -> "Furo":
        t = Tile.__decode__(data['tile'])
        if data['type'] == 'Chi':
            return Chi(t)
        elif data['type'] == 'Pon':
            return Pon(t)
        elif data['type'] == 'Kan':
            return Kan(t, data['ankan'])
        else:
            raise ValueError("invalid type: " + data['type'])

    @staticmethod
    def parse(t: Union[Sequence[Tile], str], ankan: bool = False) -> "Furo":
        if isinstance(t, str):
            if len(t) == 5 and t[0] == t[3] == '0' and t[1] == t[2]:
                t = t[1] + t[1] + t[1] + t[1] + t[4]
                ankan = True
            t = parse_tiles(t)

        if len(t) == 3:
            if t[0] == t[1] == t[2]:
                return Pon(t[0])
            else:
                if t[0].tile_type == TileType.Z or t[1].tile_type == TileType.Z or t[2].tile_type == TileType.Z:
                    raise ValueError(f"invalid tiles: {t}")
                t = sorted(t)
                if t[1] - t[0] == 1 and t[2] - t[1] == 1:
                    return Chi(t[0])
                else:
                    raise ValueError(f"invalid tiles: {t}")
        elif len(t) == 4:
            if t[0] == t[1] == t[2] == t[3]:
                return Kan(t[0], ankan)

        raise ValueError(f"invalid tiles: {t}")


@dataclass(frozen=True)
class Chi(Shuntsu, Furo):
    """
    吃
    """

    def __repr__(self):
        return f"{self.tile.num}{self.tile.num + 1}{self.tile.num + 2}{self.tile.tile_type}"


@dataclass(frozen=True)
class Pon(Kotsu, Furo):
    """
    碰
    """

    def __repr__(self):
        return f"{self.tile.num}{self.tile.num}{self.tile.num}{self.tile.tile_type}"


@dataclass(frozen=True)
class Kan(Kotsu, Furo):
    """
    杠
    """

    ankan: bool

    def __encode__(self) -> dict:
        x = Furo.__encode__(self)
        x["ankan"] = self.ankan
        return x

    def __repr__(self):
        if self.ankan:
            return f"0{self.tile.num}{self.tile.num}0{self.tile.tile_type}"
        else:
            return f"{self.tile.num}{self.tile.num}{self.tile.num}{self.tile.num}{self.tile.tile_type}"

    @property
    def tiles(self) -> List[Tile]:
        return [self.tile] * 4



import subprocess
import sys
from distutils import log
from distutils.command.clean import clean as origin_clean
from distutils.errors import DistutilsExecError
from distutils.file_util import copy_file
from distutils.util import get_platform
from pathlib import Path

from setuptools import Command, setup
from setuptools.command.build_py import build_py as origin_build_py


def run_gradle_task(root, task):
    if sys.platform == 'win32':
        gradlew = "gradlew.bat"
    else:
        gradlew = "gradlew"

    if sys.platform == 'win32':
        cmd = f"{gradlew} {task}"
    else:
        cmd = f"./{gradlew} {task}"

    log.info(f"running {cmd} (in {root})")
    call_return = subprocess.call(cmd, shell=True, cwd=root)
    if call_return != 0:
        raise DistutilsExecError(f"gradlew returned an non-zero value {call_return}")


class build_py(origin_build_py):
    def run(self):
        self.run_command('build_kt')
        return super().run()


class build_kt(Command):
    user_options = [
        ('build-lib=', 'd', "directory to \"build\" (copy) to"),
        ('kt-libraries=', None, ''),
        ('shared-location=', None, ""),
    ]

    def initialize_options(self) -> None:
        self.build_lib = None
        self.shared_location = None
        self.kt_libraries = None

    def finalize_options(self) -> None:
        self.set_undefined_options('build_py',
                                   ('build_lib', 'build_lib'))

    def get_kt_build_dir(self, lib_name, build_info):
        build_dir = Path(build_info.get("root"))
        subproject = build_info.get("subproject", None)
        if subproject is not None:
            build_dir = build_dir / subproject
        build_dir = build_dir / "build" / "bin" / "native" / "releaseShared"
        return build_dir

    def build_sharedlib(self):
        out_dir = Path(self.build_lib) / self.shared_location
        out_dir.mkdir(exist_ok=True, parents=True)

        for (lib_name, build_info) in self.kt_libraries:
            log.info("building '%s' Kotlin/Native Shared Library", lib_name)

            root = Path(build_info.get("root")).absolute()

            task = ""
            subproject = build_info.get("subproject", None)
            if subproject is not None:
                task += f":{subproject}:"
            task += "linkReleaseSharedNative"

            run_gradle_task(root, task)

            build_dir = self.get_kt_build_dir(lib_name, build_info)

            if sys.platform == 'win32':
                lib_filename = f"{lib_name}.dll"  # windows
            elif sys.platform == 'darwin':
                lib_filename = f"{lib_name}.dylib"  # macOS
            else:
                lib_filename = f"{lib_name}.so"  # unix/linux

            copy_file(str(build_dir / lib_filename), str(out_dir))

    def run(self):
        self.build_sharedlib()


class clean(origin_clean):
    def run(self) -> None:
        super().run()
        self.run_command("clean_kt")


class clean_kt(Command):
    user_options = [
        ('kt-libraries=', None, ''),
    ]

    def initialize_options(self) -> None:
        self.kt_libraries = None

    def finalize_options(self) -> None:
        self.set_undefined_options('build_kt',
                                   ('kt_libraries', 'kt_libraries'))

    def run(self):
        for (lib_name, build_info) in self.kt_libraries:
            log.info("cleaning '%s' Kotlin/Native Shared Library", lib_name)

            root = Path(build_info.get("root")).absolute()
            run_gradle_task(root, "clean")


with open("README.md", "r", encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="mahjong-utils",
    version="0.2.0a1",
    author="ssttkkl",
    author_email="huang.wen.long@hotmail.com",
    license="MIT",
    url="https://github.com/ssttkkl/mahjong-utils",
    description="立直麻将小工具",
    long_description=long_description,
    long_description_content_type="text/markdown",
    install_requires=[
        "pydantic>=1.9.0",
        "cffi>=1.15.1",
        "stringcase>=1.2.0"
    ],
    packages=[
        "mahjong_utils",
        "mahjong_utils.lib",
        "mahjong_utils.models",
        "mahjong_utils.yaku"
    ],
    package_data={"": ["*_api.i"]},
    zip_safe=False,
    options={
        "build_kt": {
            "kt_libraries": [
                ("libmahjongutils",
                 {
                     "root": "kt",
                     "subproject": "mahjong-utils-native-sharedlib"
                 })
            ],
            "shared_location": "mahjong_utils/lib"
        },
        "bdist_wheel": {
            "plat_name": get_platform()
        }
    },
    cmdclass={"build_kt": build_kt, "build_py": build_py, "clean": clean, "clean_kt": clean_kt}
)



from abc import ABC, abstractmethod
from typing import Set, Dict, Optional

from pydantic import BaseModel

from mahjong_utils.models.tile import Tile


class Shanten(BaseModel, ABC):
    shanten: int

    @abstractmethod
    def __encode__(self) -> dict:
        raise NotImplementedError()

    @classmethod
    def __decode__(cls, data: dict) -> "Shanten":
        if data['type'] == 'ShantenWithoutGot':
            return ShantenWithoutGot.__decode__(data)
        elif data['type'] == 'ShantenWithGot':
            return ShantenWithGot.__decode__(data)
        else:
            raise ValueError("invalid type: " + data['type'])


class ShantenWithoutGot(Shanten):
    advance: Set[Tile]
    advance_num: Optional[int]
    good_shape_advance: Optional[Set[Tile]]
    good_shape_advance_num: Optional[int]

    def __encode__(self) -> dict:
        return dict(
            type="ShantenWithoutGot",
            shantenNum=self.shanten,
            advance=[t.__encode__() for t in self.advance],
            advanceNum=self.advance_num,
            goodShapeAdvance=[t.__encode__() for t in self.good_shape_advance]
            if self.good_shape_advance is not None else None,
            goodShapeAdvanceNum=self.good_shape_advance_num
        )

    @classmethod
    def __decode__(cls, data: dict) -> "ShantenWithoutGot":
        return ShantenWithoutGot(
            shanten=data["shantenNum"],
            advance=set(Tile.__decode__(x) for x in data["advance"]),
            advance_num=data["advanceNum"]
            if data["advanceNum"] is not None else None,
            good_shape_advance=set(Tile.__decode__(x) for x in data["goodShapeAdvance"])
            if data["goodShapeAdvance"] is not None else None,
            good_shape_advance_num=data["goodShapeAdvanceNum"]
            if data["goodShapeAdvanceNum"] is not None else None,
        )


class ShantenWithGot(Shanten):
    discard_to_advance: Dict[Tile, ShantenWithoutGot]

    def __encode__(self) -> dict:
        return dict(
            type="ShantenWithGot",
            shantenNum=self.shanten,
            discardToAdvance=dict((k.__encode__(), v.__encode__()) for (k, v) in self.discard_to_advance.items())
        )

    @classmethod
    def __decode__(cls, data: dict) -> "ShantenWithGot":
        return ShantenWithGot(
            shanten=data["shantenNum"],
            discard_to_advance=dict(
                (Tile.__decode__(k), ShantenWithoutGot.__decode__(v))
                for (k, v) in data["discardToAdvance"].items())
        )



from enum import Enum
from typing import Optional, Sequence, Set, Dict

from pydantic import BaseModel
from stringcase import snakecase, pascalcase

from mahjong_utils.lib import libmahjongutils
from mahjong_utils.models.furo import Furo
from mahjong_utils.models.hand import Hand
from mahjong_utils.models.shanten import Shanten, ShantenWithoutGot
from mahjong_utils.models.tile import Tile


class ShantenResultType(str, Enum):
    regular = "Regular"
    chitoi = "Chitoi"
    kokushi = "Kokushi"
    union = "Union"


class ShantenResult(BaseModel):
    type: ShantenResultType
    hand: Hand
    shanten_info: Shanten
    regular: Optional["ShantenResult"]
    chitoi: Optional["ShantenResult"]
    kokushi: Optional["ShantenResult"]

    def __encode__(self) -> dict:
        return dict(
            type=pascalcase(self.type.name),
            hand=self.hand.__encode__(),
            shantenInfo=self.shanten_info.__encode__(),
            regular=self.regular.__encode__() if self.regular is not None else None,
            chitoi=self.chitoi.__encode__() if self.chitoi is not None else None,
            kokushi=self.kokushi.__encode__() if self.kokushi is not None else None,
        )

    @classmethod
    def __decode__(cls, data: dict) -> "ShantenResult":
        return ShantenResult(
            type=ShantenResultType[snakecase(data["type"])],
            hand=Hand.__decode__(data["hand"]),
            shanten_info=Shanten.__decode__(data["shantenInfo"]),
            regular=ShantenResult.__decode__(data["regular"]) if data["regular"] is not None else None,
            chitoi=ShantenResult.__decode__(data["chitoi"]) if data["chitoi"] is not None else None,
            kokushi=ShantenResult.__decode__(data["kokushi"]) if data["kokushi"] is not None else None,
        )

    @property
    def shanten(self) -> Optional[int]:
        return getattr(self.shanten_info, "shanten", None)

    @property
    def advance(self) -> Optional[Set[Tile]]:
        return getattr(self.shanten_info, "advance", None)

    @property
    def advance_num(self) -> Optional[int]:
        return getattr(self.shanten_info, "advance_num", None)

    @property
    def good_shape_advance(self) -> Optional[Set[Tile]]:
        return getattr(self.shanten_info, "good_shape_advance", None)

    @property
    def good_shape_advance_num(self) -> Optional[int]:
        return getattr(self.shanten_info, "good_shape_advance_num", None)

    @property
    def discard_to_advance(self) -> Optional[Dict[Tile, ShantenWithoutGot]]:
        return getattr(self.shanten_info, "discard_to_advance", None)

    @property
    def with_got(self) -> bool:
        return self.discard_to_advance is not None


def regular_shanten(
        tiles: Sequence[Tile],
        furo: Optional[Sequence[Furo]] = None,
        calc_advance_num: bool = True
) -> ShantenResult:
    result = libmahjongutils.call("regularShanten", {
        "tiles": [str(t) for t in tiles],
        "furo": [Furo.__encode__(fr) for fr in furo] if furo is not None else [],
        "calcAdvanceNum": calc_advance_num
    })

    return ShantenResult.__decode__(result)


def chitoi_shanten(
        tiles: Sequence[Tile],
        calc_advance_num: bool = True
) -> ShantenResult:
    result = libmahjongutils.call("chitoiShanten", {
        "tiles": [str(t) for t in tiles],
        "calcAdvanceNum": calc_advance_num
    })

    return ShantenResult.__decode__(result)


def kokushi_shanten(
        tiles: Sequence[Tile],
        calc_advance_num: bool = True
) -> ShantenResult:
    result = libmahjongutils.call("kokushiShanten", {
        "tiles": [str(t) for t in tiles],
        "calcAdvanceNum": calc_advance_num
    })

    return ShantenResult.__decode__(result)


def shanten(
        tiles: Sequence[Tile],
        furo: Optional[Sequence[Furo]] = None,
        calc_advance_num: bool = True
) -> ShantenResult:
    result = libmahjongutils.call("shanten", {
        "tiles": [str(t) for t in tiles],
        "furo": [Furo.__encode__(fr) for fr in furo] if furo is not None else [],
        "calcAdvanceNum": calc_advance_num
    })

    return ShantenResult.__decode__(result)


__all__ = ("regular_shanten",
           "chitoi_shanten",
           "kokushi_shanten",
           "shanten",
           "ShantenResult",)
